from .admin_api import (
    FeedbackAggregateViewAPI,
    FeedbackListViewAPI,
    FeedbackDetailViewAPI,
    FeedbackDeleteViewAPI,
)
from .public import (
    feedback,
    feedback_with_message,
)




from typing import Any
from django import forms
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError

from wagtail.models import Page

from . import get_feedback_model
from .backends import IPBasedFeedbackend

class AbstractFeedbackForm(forms.ModelForm):
    class Meta:
        model = get_feedback_model()
        fields = [
            "positive",
            "message",
        ]

        widgets = {
            "message": forms.Textarea(attrs={
                "placeholder": _("Message"),
                "rows": 3,
            }),
        }

        help_texts = {
            "message": _("What would you like to see improved?"),
        }
    
    def __init__(self, *args, request: HttpRequest = None, page: Page = None, requires_message: bool = True, **kwargs):
        self.request = request
        self.page = page
        self.requires_message = requires_message
        super().__init__(*args, **kwargs)
        if self.requires_message:
            del self.fields["positive"]
            self.fields["message"].required = True
        else:
            del self.fields["message"]
    
    def clean(self):
        cleaned = super().clean()
        if self.requires_message:
            message = cleaned.get("message", "")
            if not message.strip():
                raise ValidationError(_("You must provide a message."))
        
        return cleaned

class FeedbackForm(AbstractFeedbackForm):
    def save(self, commit: bool = True) -> Any:
        instance = super().save(False)
        if not self.request and not instance.ip_address:
            raise RuntimeError("A request must be passed to the form when the instance does not have an IP address.")

        elif self.request and not instance.ip_address:
            instance.ip_address = IPBasedFeedbackend.ip_address(self.request)

        if commit:
            instance.save()

        return instance
    





from typing import TYPE_CHECKING, Union
from django.http import HttpRequest
from django.utils.module_loading import import_string
from wagtail.models import Page

from .base import Feedbackend, get_feedback_backend

if TYPE_CHECKING:
    from feedback.models import AbstractFeedback
    from feedback.forms import FeedbackForm
    from feedback.mixins import FeedbackendPageMixin



class PageBasedFeedbackend(Feedbackend):
    def __init__(self, options: dict = None):
        super().__init__(options)
        backup_backend = self.options.get("BACKUP_BACKEND", None)
        backup_backend_options = self.options.get("BACKUP_BACKEND_OPTIONS", {})

        if backup_backend:
            self.backup = get_feedback_backend(
                klass=backup_backend,
                options=backup_backend_options,
            )
        else:
            self.backup = None


    def is_duplicate(self, request: HttpRequest, page: Union[Page, FeedbackendPageMixin], form: "FeedbackForm", exists: bool = False) -> bool:
        if hasattr(page, "check_for_feedback_duplicate"):
            return page.check_for_feedback_duplicate(request, form, exists=exists)
        
        if self.backup:
            return self.backup.is_duplicate(request, page, form, exists=exists)
        
        return False


    def end_check(self, request: HttpRequest, page: Union[Page, FeedbackendPageMixin], form: "FeedbackForm", instance: "AbstractFeedback", exists: bool = False) -> None:
        if hasattr(page, "end_feedback_check"):
            page.end_feedback_check(request, form, instance, exists=exists)
            return

        if self.backup:
            self.backup.end_check(request, page, form, instance, exists=exists)
            return




from typing import TYPE_CHECKING
from django.http import HttpRequest
from wagtail.models import Page
from ..options import (
    IS_PROXIED,
)
from .base import Feedbackend
from .. import get_feedback_model

if TYPE_CHECKING:
    from feedback.models import AbstractFeedback
    from feedback.forms import FeedbackForm



Feedback = get_feedback_model()


class IPBasedFeedbackend(Feedbackend):
    def is_duplicate(self, request: HttpRequest, page: Page, form: "FeedbackForm", exists: bool = False) -> bool:
        if not request:
            raise RuntimeError("A request must be passed to the form when the instance does not have an IP address.")

        ip_address = self.ip_address(request)
        feedback_qs = Feedback.objects.filter(
            ip_address=ip_address,
            page=page,
        )

        if exists:
            feedback_qs = feedback_qs.filter(
                message__isnull=False,
            )
        else:
            feedback_qs = feedback_qs.filter(
                message__isnull=True,
            )

        return feedback_qs.exists()

    
    def end_check(self, request: HttpRequest, page: Page, form: "FeedbackForm", instance: "AbstractFeedback", exists: bool = False) -> None:
        pass

    @staticmethod
    def ip_address(request: HttpRequest):
        if IS_PROXIED:
            addr: str = request.META.get('HTTP_X_FORWARDED_FOR', None)
            if addr:
                return addr.split(',')[-1].strip()
        else:
            return request.META.get('REMOTE_ADDR', None)




from typing import Literal as L, Self, Type, Union, TYPE_CHECKING
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.utils.functional import classproperty
from django.utils.module_loading import import_string

from wagtail.models import Page
from wagtail.admin.panels import (
    FieldPanel,
    TabbedInterface,
    ObjectList,
)

from .options import (
    FEEDBACK_FORM_CLASS,
    FEEDBACK_FILTER_CLASS,
)

if TYPE_CHECKING:
    from feedback.filters import AbstractFeedbackFilter
    from feedback.forms import AbstractFeedbackForm


class FeedbackQuerySet(models.QuerySet):
    default_period: Union[L['hour'], L['date'], L["month"], L["year"]] = "hour"

    def positive(self):
        return self.filter(positive=True)
    
    def negative(self):
        return self.filter(positive=False)
    
    def for_page(self, page):
        return self.filter(page=page)
    
    def percentages(self):
        return self.annotate(
            total=models.Count('id'),
            positive_count=models.Count('id', filter=models.Q(positive=True)),
            negative_count=models.Count('id', filter=models.Q(positive=False)),
        ).annotate(
            positive_percentage=models.ExpressionWrapper(
                models.F('positive_count') * 100.0 / models.F('total'),
                output_field=models.FloatField()
            ),
            negative_percentage=models.ExpressionWrapper(
                100.0 - models.F('positive_percentage'),
                output_field=models.FloatField()
            ),
        )
    
    def aggregate_percentage(self, date_arg: Union[L['hour'], L['date'], L["month"], L["year"]] = None, extra_values_args: list[str] = None):
        if extra_values_args is None:
            extra_values_args = []

        filter_by = date_arg

        if filter_by is None:
            filter_by = self.default_period

        values_args = []
        hours = filter_by == "hour"
        if hours:
            filter_by = "date"
            values_args.append(f"created_at__{filter_by}")
            values_args.append("created_at__hour")
        else:
            values_args.append(f"created_at__{filter_by}")
        
        values_args.extend(extra_values_args)

        qs = self.values(*values_args).percentages()

        if hours:
            qs = qs.order_by(f"-created_at__{filter_by}", "-created_at__hour")
        else:
            qs = qs.order_by(f"-created_at__{filter_by}")

        setattr(qs, "period", filter_by)
        setattr(qs, "hours", hours)
        return qs


class AbstractFeedback(models.Model):
    FEEDBACK_FORM_CLASS = FEEDBACK_FORM_CLASS
    FEEDBACK_FILTER_CLASS = FEEDBACK_FILTER_CLASS

    panel_tabs: list[tuple[str, str]] = [
        ("content_panels", _("Content")),
        ("metadata_panels", _("Metadata")),
    ]

    positive = models.BooleanField(
        blank=False,
        null=False,
        verbose_name=_("Positive"),
        help_text=_("Whether the feedback is positive or negative."),
    )
    message = models.TextField(
        blank=False,
        null=True,
        verbose_name=_("Message"),
        help_text=_("The message of the feedback."),
    )
    page = models.ForeignKey(
        "wagtailcore.Page",
        on_delete=models.CASCADE,
        blank=False,
        null=False,
        related_name="+",
        verbose_name=_("Page"),
        help_text=_("The page the feedback is for."),
    )
    created_at = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_("Created At"),
        help_text=_("The time the feedback was created."),
    )

    content_panels = [
        FieldPanel("positive"),
        FieldPanel("message"),
    ]

    metadata_panels = [
        FieldPanel("page"),
        FieldPanel("created_at"),
    ]

    objects: FeedbackQuerySet = FeedbackQuerySet.as_manager()

    class Meta:
        abstract = True

    @classmethod
    def serialize(cls, instance: Self):
        return {
            "id": instance.pk,
            "page": instance.page_id,
            "message": instance.message,
            "created_at": instance.created_at,
            "urls": {
                "list":      reverse("feedback_api_list"),
                "page_list": reverse("page_feedback_api", kwargs={"page_pk": instance.page_id}),
                "view":      reverse("feedback_api_detail", kwargs={"page_pk": instance.page_id, "pk": instance.pk}),
                "detail":    reverse("feedback_api_detail", kwargs={"pk": instance.pk}),
                "delete":    reverse("feedback_api_delete", kwargs={"pk": instance.pk}),
            },
        }
    
    @classmethod
    def get_filter_class(cls) -> Type["AbstractFeedbackFilter"]:
        filter_class = cls.FEEDBACK_FILTER_CLASS
        if isinstance(filter_class, str):
            filter_class = import_string(filter_class)
        return filter_class

    @classmethod
    def get_form_class(cls) -> Type["AbstractFeedbackForm"]:
        form_class = cls.FEEDBACK_FORM_CLASS
        if isinstance(form_class, str):
            form_class = import_string(form_class)
        return form_class

    def __str__(self):
        if self.message and len(self.message) > 50:
            return f"{self.message[:50]}... ({self.page.title})"
        elif self.message:
            return f"{self.message} ({self.page.title})"
        else:
            message = _("Positive") if self.positive else _("Negative")
            return f"{message} ({self.page.title})"

    @property
    def is_positive(self):
        return self.positive

    @property
    def is_negative(self):
        return not self.positive
    
    @property
    def negative(self):
        return not self.positive
    
    @is_positive.setter
    def is_positive(self, value):
        self.positive = value

    @is_negative.setter
    def is_negative(self, value):
        self.positive = not value

    @negative.setter
    def negative(self, value):
        self.positive = not value


    # @property
    # def is_neutral(self):
    #     return (
    #         # Allow for a 10% margin of error
    #         self.is_positive + 10 > self.is_negative
    #         and self.is_positive - 10 < self.is_negative
    #     )\
    #     or int(self.is_positive) == int(self.is_negative)
    
    @classmethod
    def edit_handler_kwargs(cls):
        return {}

    @classproperty
    def edit_handler(cls):
        tabs = []

        for panel_attr, heading in cls.panel_tabs:
            tabs.append(
                ObjectList(getattr(cls, panel_attr), heading=heading),
            )

        return TabbedInterface(tabs, **cls.edit_handler_kwargs())


class Feedback(AbstractFeedback):
    ip_address = models.GenericIPAddressField(
        blank=True,
        null=True,
        verbose_name=_("IP Address"),
        help_text=_("The IP address of the feedback."),
    )

    metadata_panels = AbstractFeedback.metadata_panels + [
        FieldPanel("ip_address"),
    ]

    objects: FeedbackQuerySet = FeedbackQuerySet.as_manager()

    class Meta:
        verbose_name = _("Feedback")
        verbose_name_plural = _("Feedback")
        ordering = ["-created_at"]

    def serialize(cls, instance: Self):
        return super().serialize(instance) | {
            "ip_address": instance.ip_address,
        }

    


from typing import TYPE_CHECKING
from django.http import HttpRequest
from wagtail.models import Page

from .base import Feedbackend

if TYPE_CHECKING:
    from feedback.models import AbstractFeedback
    from feedback.forms import FeedbackForm



class SessionBasedFeedbackend(Feedbackend):
    def __init__(self, options: dict = None):
        super().__init__(options)
        self.format_key: str = self.options.get("FORMAT_KEY", "user-feedback-{page.pk}")
        self.format_exists_key: str = self.options.get("FORMAT_EXISTS_KEY", "user-feedback-has-message-{page.pk}")
        

    def is_duplicate(self, request: HttpRequest, page: Page, form: "FeedbackForm", exists: bool = False) -> bool:
        if not request:
            raise RuntimeError("A request must be passed to the form when the instance does not have an IP address.")

        if exists:
            key = self.format_exists_key.format(page=page)
        else:
            key = self.format_key.format(page=page)

        return key in request.session

    def end_check(self, request: HttpRequest, page: Page, form: "FeedbackForm", instance: "AbstractFeedback", exists: bool = False) -> None:
        if exists:
            key = self.format_exists_key.format(page=page)
        else:
            key = self.format_key.format(page=page)
        request.session[key] = True
        request.session.modified = True





from typing import Union, Literal
from django.template import library, loader
from django.utils.safestring import mark_safe
from django.templatetags.static import static
from django.template.defaultfilters import date as date_filter
from django.utils import timezone
import datetime
from wagtail.models import PAGE_TEMPLATE_VAR

register = library.Library()

FEEDBACK_TEMPLATE = "feedback/happy-sad.html"
FEEDBACK_CSS = "feedback/css/feedback.css"
FEEDBACK_JS = "feedback/js/feedback.js"

@register.simple_tag(takes_context=True, name="feedback")
def feedback(context, page = None):
    request = context["request"]

    if page is None and PAGE_TEMPLATE_VAR not in context:
        raise RuntimeError("No page found in context")
    
    if page is None:
        page = context[PAGE_TEMPLATE_VAR]

    context = {
        "page": page,
        "request": request,
    }

    return mark_safe(loader.render_to_string(
        FEEDBACK_TEMPLATE,
        context,
        request=request,
    ))

@register.simple_tag(name="feedback_css")
def feedback_css():
    return mark_safe(f"<link rel='stylesheet' href='{static(FEEDBACK_CSS)}'>")

@register.simple_tag(name="feedback_js")
def feedback_js():
    return mark_safe(f"<script src='{static(FEEDBACK_JS)}'></script>")

@register.simple_tag
def get_proper_elided_page_range(p, number, on_each_side=3, on_ends=2):
    return p.get_elided_page_range(number=number, on_each_side=on_each_side, on_ends=on_ends)

class _impossible:
    def __bool__(self):
        return False
    
@register.simple_tag(name="page_params", takes_context=True)
def do_page_params(context, page_param: str, page_index: int):
    request = context["request"]
    query = request.GET.copy()
    query[page_param] = page_index
    return query.urlencode()

@register.simple_tag(name="get_date")
def do_date(obj, default=_impossible, **formats):

    for field, format in formats.items():
        if isinstance(obj, dict):
            value = obj.get(field, default)
        else:
            value = getattr(obj, field, default)

        if value is not _impossible:
            break

    if value is _impossible and default is _impossible:
        raise RuntimeError("No date found")
    
    if value is _impossible:
        return default
    
    if not isinstance(value, (datetime.date, datetime.datetime, str)):
        return value

    return date_filter(value, format)

@register.simple_tag(name="format_date")
def do_format_date(year = None, month = None, day = None, hour = None, minute = None, second = None, absolute = None, period: Union[Literal['hour'], Literal['date'], Literal['month'], Literal['year']] = "hour"):
    if year is None and month is None and day is None and absolute is None:
        raise RuntimeError("No date found")
    
    if not absolute:
        absolute = timezone.now() 
    
    if not year and year is not 0:
        year = absolute.year

    if not month and month is not 0:
        month = absolute.month

    if not day and day is not 0:
        day = absolute.day

    if not hour and hour is not 0:
        hour = getattr(absolute, "hour", 0)

    if not minute and minute is not 0:
        minute = getattr(absolute, "minute", 0)

    if not second and second is not 0:
        second = getattr(absolute, "second", 0)

    try:
        year, month, day, hour, minute, second =\
            map(int, (year, month, day, hour, minute, second))
    except ValueError:
        return date_filter(absolute, "j F Y H:i")

    match period:
        case "hour":
            value = datetime.datetime(year, month, day, hour, minute, second)
            return date_filter(value, "j F Y H:i")
        case "date":
            value = datetime.date(year, month, day)
            return date_filter(value, "j F Y")
        case "month":
            value = datetime.date(year, month, day)
            return date_filter(value, "F Y")
        case "year":
            value = datetime.date(year, month, day)
            return date_filter(value, "Y")
        case _:
            return absolute







from django.urls import reverse
from wagtail.admin.panels import Panel

from .templatetags.feedback import (
    FEEDBACK_CSS,
    FEEDBACK_JS,
)

from . import get_feedback_model

Feedback = get_feedback_model()

class FeedbackPanel(Panel):

    def __init__(self, *args, visible_by_default = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.visible_by_default = visible_by_default

    class BoundPanel(Panel.BoundPanel):
        template_name = "feedback/panels/feedback_panel.html"
        panel: "FeedbackPanel"

        class Media:
            js = [
                FEEDBACK_JS,
                "wagtailadmin/js/date-time-chooser.js"
            ]
            css = {
                "all": [
                    FEEDBACK_CSS,
                ]
            }

        def is_shown(self):
            shown = super().is_shown() and self.request.user.has_perm("feedback.view_feedback")
            if not shown:
                return False
            
            if self.instance is None:
                return False
            
            if not self.instance.pk:
                return False
            
            return Feedback.objects.filter(page=self.instance).exists()

        def get_context_data(self, parent_context=None):
            context = super().get_context_data(parent_context)
            context.update(
                self.panel.get_panel_context(self.instance)
            )
            return context
        
    @classmethod
    def get_panel_context(cls, instance, **kwargs):
        return {
            "page": instance,
            "panel_id": f"feedback-panel-{instance.pk}",
            "list_url": reverse("page_feedback_api", kwargs={"page_pk": instance.pk}),
            "chart_url": reverse("page_feedback_api_chart", kwargs={"page_pk": instance.pk}),
        }



from typing import Literal as L
from django.conf import settings


FEEDBACK_MODEL_NAME = getattr(settings, "FEEDBACK_MODEL_NAME", "feedback.Feedback")
FEEDBACK_FORM_CLASS = getattr(settings, "FEEDBACK_FORM_CLASS", "feedback.forms.FeedbackForm")
FEEDBACK_FILTER_CLASS = getattr(settings, "FEEDBACK_FILTER_CLASS", "feedback.filters.AbstractFeedbackFilter")
FEEDBACK_BACKEND = getattr(settings, "FEEDBACK_BACKEND", {
    "CLASS": "feedback.backends.SessionBasedFeedbackend",
    "OPTIONS": {
        # ...
    }
})
IS_PROXIED = getattr(settings, "USE_X_FORWARDED_HOST", False)



from typing import Callable, Type
from django.forms import ValidationError
from django.shortcuts import (
    get_object_or_404,
)
from django.utils.translation import gettext_lazy as _
from django.http import (
    HttpRequest,
    HttpResponseNotAllowed,
)
from wagtail.models import (
    Page,
)
from ..backends import (
    get_feedback_backend,
)
from .. import (
    get_feedback_model,
)
from .utils import (
    redirect_or_respond,
    error,
)

from wagtail import hooks


Feedback = get_feedback_model()
FeedbackForm = Feedback.get_form_class()


_BeforeFunc = Type[Callable[[HttpRequest, Type[FeedbackForm]], None]]
_AfterFunc = Type[Callable[[HttpRequest, Type[Feedback]], None]]


before_feedback_form_valid         = "before_feedback_form_valid"
after_feedback_form_valid          = "after_feedback_form_valid"
before_feedback_message_form_valid = "before_feedback_message_form_valid"
after_feedback_message_form_valid  = "after_feedback_message_form_valid"


# Create your views here.
def feedback(request, *args, **kwargs):        
    if not request.method == "POST":
        return HttpResponseNotAllowed(["POST"])
    
    template = "feedback/form.html"

    page_qs = Page.objects.live().public().specific()
    page: Page = get_object_or_404(page_qs, pk=kwargs.get("page_pk", None))
    form = Feedback.get_form_class()(
        request.POST,
        request=request,
        requires_message=False,
    )

    backend = get_feedback_backend()
    if backend.is_duplicate(request, page, form, exists=False):
        return error(
            request, 
            _("You have already submitted feedback for this page."), 
            WRAPPER="feedback/wrapper.html", 
            to=page.get_url(request),
        )

    try:
        hks: list[_BeforeFunc] = hooks.get_hooks(before_feedback_form_valid)
        for fn in hks:
            fn(request, form)
    except ValidationError as e:
        form.add_error(None, e)

    valid = form.is_valid()

    if valid:
        form.instance.page = page
        form.instance = form.save()

        hks: list[_AfterFunc] = hooks.get_hooks(after_feedback_form_valid)
        for fn in hks:
            fn(request, form.instance)

        backend.end_check(request, page, form, form.instance, exists=False)

        # If the feedback is positive and it is allowed, or if it is negative
        # then show the message form.
        if hasattr(page, "allow_feedback_message_on_positive")\
            and not page.allow_feedback_message_on_positive()\
            and form.instance.positive\
            or not form.instance.positive:

            form = FeedbackForm(
                request=request,
                page=page,
                requires_message=True,
                instance=form.instance,
            )
        else:
            template = "feedback/thanks.html"
    else:
        template = "feedback/happy-sad.html"

    context = page.get_context(request, *args, **kwargs)
    context["form"] = form
    context["feedback"] = form.instance

    if form.errors:
        context["errors"] = form.errors

    return redirect_or_respond(
        request,
        page.get_url(request),
        template,
        context=context,
        message=_("Thank you for your feedback."),
    )


def feedback_with_message(request, *args, **kwargs):
    if not request.method == "POST":
        return HttpResponseNotAllowed(["POST"])
    
    template = "feedback/form.html"

    page_qs = Page.objects.live().public().specific()
    page: Page = get_object_or_404(page_qs, pk=kwargs.get("page_pk", None))
    feedback = get_object_or_404(Feedback, pk=kwargs.get("pk", None))

    if hasattr(page, "allow_feedback_message_on_positive") \
        and not page.allow_feedback_message_on_positive() \
        and feedback.positive:
        return error(
            request, 
            _("Feedback messages are not allowed on positive feedback."), 
            WRAPPER="feedback/wrapper.html", 
            to=page.get_url(request),
        )

    form = FeedbackForm(
        request.POST,
        request=request,
        page=page,
        requires_message=True,
        instance=feedback,
    )

    backend = get_feedback_backend()
    if backend.is_duplicate(request, page, form, exists=True):
        return error(
            request, 
            _("You have already submitted feedback for this page."), 
            WRAPPER="feedback/wrapper.html", 
            to=page.get_url(request),
        )

    try:
        hks: list[_BeforeFunc] = hooks.get_hooks(before_feedback_message_form_valid)
        for fn in hks:
            fn(request, form)
    except ValidationError as e:
        form.add_error(None, e)

    if form.is_valid():
        form.instance.page = page
        form.instance = form.save()
        template = "feedback/thanks.html"

        backend.end_check(request, page, form, form.instance, exists=True)

        hks: list[_AfterFunc] = hooks.get_hooks(after_feedback_message_form_valid)
        for fn in hks:
            fn(request, form.instance)

    context = page.get_context(request, *args, **kwargs)
    context["form"] = form
    context["feedback"] = form.instance

    return redirect_or_respond(
        request,
        page.get_url(request),
        template,
        context=context,
    )

